from .make_request import *
from urllib.parse import urljoin


# ============================================================
# URL + HTTP helpers
# ============================================================

def normalize(domain: str) -> str:
    return domain.rstrip("/")


def join(domain: str, path: str) -> str:
    domain = normalize(domain)
    path = path.lstrip("/")
    return f"{domain}/{path}"


def is_valid_endpoint(url: str) -> bool:
    """Return True if the URL responds with anything < 500-level error."""
    try:
        r = requests.get(url, timeout=3)
        return r.status_code < 500
    except Exception:
        return False


def fetch_json(url: str):
    """Return JSON if response==200, else None."""
    try:
        r = requests.get(url, timeout=3)
        if r.status_code == 200:
            return r.json()
    except Exception:
        pass
    return None


# ============================================================
# Prefix discovery
# ============================================================

def get_prefixes(domain: str, prefix: str = None):
    """
    Returns a list of full prefix URLs:
       - If prefix is given, try only that prefix
       - Otherwise, query /prefixes
    """
    domain = normalize(domain)

    # Direct prefix override
    if prefix:
        test_url = join(domain, prefix)
        js = fetch_json(test_url)
        if js is not None:
            return [test_url]     # prefix is valid
        # else continue to auto-discovery

    # Auto-discover prefixes from /prefixes
    prefixes_url = join(domain, "prefixes")
    pref_list = fetch_json(prefixes_url)
    if not pref_list:
        return []

    # Map returned prefix strings → absolute URLs
    return [join(domain, p) for p in pref_list]


# ============================================================
# Endpoint discovery
# ============================================================

def get_endpoint_urls(domain: str, prefix: str = None):
    """
    Returns URLs like:
        /utilities/endpoints
        /math/endpoints
    for prefixes that respond correctly.
    """
    endpoints = []
    for prefix_url in get_prefixes(domain, prefix=prefix):
        ep_url = f"{prefix_url}/endpoints"
        js = fetch_json(ep_url)
        if js is not None:
            endpoints.append(ep_url)
    return endpoints


def get_endpoints(domain: str, prefix: str = None, *args, **kwargs):
    """
    Returns endpoint entries from each prefix-level /endpoints.
    Expected structure: [["/utilities/load_from_json", "GET,POST"], ...]
    """
    all_eps = []
    for ep_url in get_endpoint_urls(domain, prefix):
        result = postRequest(ep_url, **kwargs) if kwargs else postRequest(ep_url)
        if isinstance(result, list):
            all_eps.extend(result)
    return all_eps


# ============================================================
# Endpoint lookup
# ============================================================

def get_endpoint(name: str, domain: str, prefix: str = None, *args, **kwargs):
    """
    Locate endpoint containing the substring `name`.
    Returns full URL: domain + /prefix/.../name
    """
    domain = normalize(domain)
    endpoints = get_endpoints(domain, prefix, *args, **kwargs)

    matches = [e for e in endpoints if name in e[0]]
    
    if matches:
        return join(domain, matches[0][0])

    return None


def check_endpoint(domain: str, endpoint: str, prefix: str = None):
    """
    First try direct:   domain/endpoint
    Fallback:           search /prefix/endpoints
    """
    domain = normalize(domain)

    direct = join(domain, endpoint)
    if is_valid_endpoint(direct):
        return direct

    found = get_endpoint(endpoint, domain=domain, prefix=prefix)
    return found

def make_endpoint_call(*args,domain,endpoint,prefix=None,**kwargs):
    url = get_endpoint(endpoint, domain=domain, prefix=prefix)
    if url:
        response = makeRequest(url,*args,**kwargs)
        return response



import logging
import os
import requests

from .request_utils import *


logging.basicConfig(level=logging.WARNING)


def close_request_files(files):
    if not files or not isinstance(files, dict):
        return

    for value in files.values():
        if hasattr(value, "close"):
            value.close()
            continue

        if isinstance(value, tuple):
            for part in value:
                if hasattr(part, "close"):
                    part.close()


def make_request(
    url,
    data=None,
    json_data=None,
    headers=None,
    get_post=None,
    endpoint=None,
    files=None,
    timeout=600,
    status_code=False,
    retry_after=False,
    raw_response=False,
    response_result=None,
    load_nested_json=True,
    auth=None,
    logger=True,
):
    data = data or {}

    values = get_values_js(
        url=url,
        endpoint=endpoint,
        data=data,
        headers=headers,
        auth=auth,
        files=files,
        timeout=timeout,
    )

    if json_data and not files:
        values.pop("data", None)
        values["json"] = json_data

    method = str(
        get_post or ("POST" if files or data or json_data else "GET")
    ).upper()

    try:
        response = requests.request(method=method, **values)
    except Exception as e:
        logging.error(f"Request failed: {e}")
        raise

    return get_response(
        response,
        raw_response=raw_response,
        response_result=response_result,
        load_nested_json=load_nested_json,
    )


def postRequest(
    url,
    data=None,
    headers=None,
    endpoint=None,
    request_file_path=None,
    files=None,
    file_field_name="files",
    filename=None,
    mime_type=None,
    timeout=600,
    status_code=False,
    retry_after=False,
    raw_response=False,
    response_result=None,
    load_nested_json=True,
    auth=None,
    **kwargs,
):
    owns_files = False

    if request_file_path:
        files = get_request_file(
            request_file_path,
            field_name=file_field_name,
            filename=filename,
            mime_type=mime_type,
        )
        owns_files = True

    elif files and isinstance(files, str):
        files = get_request_file(
            files,
            field_name=file_field_name,
            filename=filename,
            mime_type=mime_type,
        )
        owns_files = True

    data = data or kwargs or {}

    try:
        return make_request(
            url=url,
            data=data,
            headers=headers,
            endpoint=endpoint,
            get_post="POST",
            files=files,
            timeout=timeout,
            status_code=status_code,
            retry_after=retry_after,
            raw_response=raw_response,
            response_result=response_result,
            load_nested_json=load_nested_json,
            auth=auth,
        )
    finally:
        if owns_files:
            close_request_files(files)


def getRequest(
    url,
    data=None,
    headers=None,
    endpoint=None,
    request_file_path=None,
    files=None,
    file_field_name="files",
    timeout=600,
    status_code=False,
    retry_after=False,
    raw_response=False,
    response_result=None,
    load_nested_json=True,
    auth=None,
    **kwargs,
):
    owns_files = False

    if request_file_path:
        files = get_request_file(request_file_path, field_name=file_field_name)
        owns_files = True

    elif files and isinstance(files, str):
        files = get_request_file(files, field_name=file_field_name)
        owns_files = True

    data = data or kwargs or {}

    try:
        return make_request(
            url=url,
            data=data,
            headers=headers,
            endpoint=endpoint,
            get_post="GET",
            files=files,
            timeout=timeout,
            status_code=status_code,
            retry_after=retry_after,
            raw_response=raw_response,
            response_result=response_result,
            load_nested_json=load_nested_json,
            auth=auth,
        )
    finally:
        if owns_files:
            close_request_files(files)


def makeRequest(
    url,
    *args,
    data=None,
    headers=None,
    endpoint=None,
    get_post=None,
    request_file_path=None,
    files=None,
    file_field_name="files",
    timeout=600,
    status_code=False,
    retry_after=False,
    raw_response=False,
    response_result=None,
    load_nested_json=True,
    auth=None,
    **kwargs,
):
    owns_files = False

    if request_file_path:
        files = get_request_file(request_file_path, field_name=file_field_name)
        owns_files = True

    elif files and isinstance(files, str):
        files = get_request_file(files, field_name=file_field_name)
        owns_files = True

    data = data or kwargs or {}

    if not isinstance(data, dict):
        data = {"args": make_list(data or [])}

    data["args"] = make_list(data.get("args") or []) + list(args)

    try:
        return make_request(
            url=url,
            data=data,
            headers=headers,
            endpoint=endpoint,
            get_post=get_post,
            files=files,
            timeout=timeout,
            status_code=status_code,
            retry_after=retry_after,
            raw_response=raw_response,
            response_result=response_result,
            load_nested_json=load_nested_json,
            auth=auth,
        )
    finally:
        if owns_files:
            close_request_files(files)


def getRpcRequest(
    url,
    method=None,
    params=None,
    jsonrpc=None,
    id=None,
    headers=None,
    endpoint=None,
    status_code=False,
    retry_after=False,
    raw_response=False,
    response_result=None,
    load_nested_json=True,
    auth=None,
):
    data = getRpcData(
        method=method,
        params=params,
        jsonrpc=jsonrpc,
        id=id,
    )

    return getRequest(
        url,
        data=data,
        headers=headers,
        endpoint=endpoint,
        status_code=status_code,
        retry_after=retry_after,
        raw_response=raw_response,
        response_result=response_result,
        load_nested_json=load_nested_json,
        auth=auth,
    )


def postRpcRequest(
    url,
    method=None,
    params=None,
    jsonrpc=None,
    id=None,
    headers=None,
    endpoint=None,
    status_code=False,
    retry_after=False,
    raw_response=False,
    response_result=None,
    load_nested_json=True,
    auth=None,
):
    data = getRpcData(
        method=method,
        params=params,
        jsonrpc=jsonrpc,
        id=id,
    )

    return make_request(
        url=url,
        data=data,
        headers=headers,
        endpoint=endpoint,
        get_post="POST",
        status_code=status_code,
        retry_after=retry_after,
        raw_response=raw_response,
        response_result=response_result,
        load_nested_json=load_nested_json,
        auth=auth,
    )



import json
import os
import requests
import aiohttp
import asyncio

from flask import jsonify
from abstract_essentials import make_list, eatAll, get_mime_type


def get_headers(is_json=True):
    if is_json:
        return {"Content-Type": "application/json"}
    return {}


def ensure_json(data):
    if isinstance(data, str):
        try:
            json.loads(data)
            return data
        except ValueError:
            pass
    return json.dumps(data)


def stripit(string, chars=[]):
    string = string or ""
    for char in make_list(chars):
        string = string.strip(char)
    return string


def make_endpoint(endpoint):
    return eatAll(endpoint, ["/"])


def make_url(url):
    return eatAll(url, ["/"])


def get_url(url, endpoint=None):
    url = eatAll(url, ["/"])
    endpoint = eatAll(endpoint, ["/"])
    return eatAll(f"{url}/{endpoint}", ["/"])


def get_values_js(
    url=None,
    data=None,
    headers=None,
    endpoint=None,
    auth=None,
    files=None,
    timeout=600,
):
    if endpoint:
        url = get_url(url, endpoint=endpoint)

    values = {
        "url": url,
        "timeout": timeout,
    }

    if auth:
        values["auth"] = auth

    if files:
        values["files"] = files
        values["data"] = data or {}

        # Never manually set Content-Type for multipart requests.
        # requests will generate the multipart boundary.
        if headers:
            values["headers"] = {
                k: v
                for k, v in headers.items()
                if k.lower() != "content-type"
            }

        return values

    values["json"] = data or {}

    if headers:
        values["headers"] = headers
    else:
        values["headers"] = get_headers(is_json=True)

    return values


def get_text_response(response):
    try:
        return response.text
    except Exception:
        return None


def load_inner_json(data):
    if isinstance(data, str):
        try:
            return load_inner_json(json.loads(data))
        except (ValueError, TypeError):
            return data

    if isinstance(data, dict):
        return {key: load_inner_json(value) for key, value in data.items()}

    if isinstance(data, list):
        return [load_inner_json(item) for item in data]

    return data


def get_status_code(response):
    try:
        return response.status_code
    except Exception as e:
        print(f"Could not get status code: {e}")
        return None


def get_retry_after(response):
    try:
        return response.headers.get("Retry-After")
    except Exception as e:
        print(f"Could not get Retry-After: {e}")
        return None


def get_json_response(
    response=None,
    response_result=None,
    load_nested_json=True,
    status_code=None,
    value=None,
):
    if response is None and response_result is None and (status_code or value):
        return jsonify({"result": value}), status_code

    response_result = response_result or "result"

    try:
        try:
            response_json = response.json()
        except Exception:
            response_json = response

        if load_nested_json:
            response_json = load_inner_json(response_json)

        if isinstance(response_json, dict):
            response_json = response_json.get(response_result, response_json)

        if response_json is not None:
            return response_json

    except Exception as e:
        print(e)
        return response_result

    return response_result


def get_response(
    response,
    response_result=None,
    raw_response=False,
    load_nested_json=True,
):
    if raw_response:
        return response

    json_response = get_json_response(
        response,
        response_result=response_result,
        load_nested_json=load_nested_json,
    )

    if json_response is not None:
        return json_response

    text_response = get_text_response(response)

    if text_response:
        return text_response

    return response


def getRpcData(method=None, params=None, jsonrpc=None, id=None):
    return {
        "jsonrpc": jsonrpc or "2.0",
        "id": id or 1,
        "method": method,
        "params": params or [],
    }


def get_request_file(file_path, field_name="files", filename=None, mime_type=None):
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

    filename = filename or os.path.basename(file_path)
    mime_type = mime_type or get_mime_type(file_path) or "application/octet-stream"

    file_handle = open(file_path, "rb")

    return {
        field_name: (
            filename,
            file_handle,
            mime_type,
        )
    }
